#!/bin/bash

set -e
STASHDIR=~/.atticman

fatal() {
	echo "FATAL: $1" >&2
	exit 1
}

require_root() {
	if [[ "$UID" != "0" ]] ; then
		fatal "This operation requires root!"
	fi
}

install_pip() {
	require_root

	if ! dpkg -s whiptail > /dev/null 2>&1 ; then
		apt-get -y install whiptail
	fi
	
	if ! dpkg -s build-essential > /dev/null 2>&1 ; then
		apt-get -y install build-essential
	fi
	
	if ! dpkg -s python3-pip > /dev/null 2>&1 ; then
		apt-get -y install python3-pip
	fi
	
	if ! dpkg -s libacl1-dev > /dev/null 2>&1 ; then
		apt-get -y install libacl1-dev
	fi
}

check_pip() {
	# Ensure attic dependencies are installed
	
	if ! dpkg -s whiptail > /dev/null 2>&1 ; then
		fatal "Missing dependencies - please run --dependencies as root"
	fi	
	if ! dpkg -s build-essential > /dev/null 2>&1 ; then
		fatal "Missing dependencies - please run --dependencies as root"
	fi
	if ! dpkg -s python3-pip > /dev/null 2>&1 ; then
		fatal "Missing dependencies - please run --dependencies as root"
	fi
	if ! dpkg -s libacl1-dev > /dev/null 2>&1 ; then
		fatal "Missing dependencies - please run --dependencies as root"
	fi
}

install_server() {
	
	check_pip
	
	# Create user account
	
	echo "Creating system user..."
	
	if [[ -z "$ATTICINST_USERNAME" ]] ; then
		USERNAME="AB$(date +%s)"
	else
		USERNAME="$ATTICINST_USERNAME"
	fi
	
	PASSWORD=$( ( dd if=/dev/urandom count=1 bs=256 2>/dev/null) | md5sum -b | cut -d' ' -f1)
	HOMEDIR=/home/${USERNAME}
	
	useradd "$USERNAME"
	echo "${USERNAME}:${PASSWORD}" | chpasswd
	
	# Create folder structure
	
	BACKUPDATA_DIR=${HOMEDIR}/backupdata
	mkdir -p "$BACKUPDATA_DIR"
	
	SSH_AUTHKEYS=${HOMEDIR}/.ssh/authorized_keys
	mkdir -p "${HOMEDIR}/.ssh"
	
	# Read ssh pubkey
	
	cat <<EOD > ${SSH_AUTHKEYS}
# Paste in the public key from the client...
command="${HOMEDIR}/.local/bin/attic serve",no-port-forwarding,no-X11-forwarding,no-agent-forwarding ssh-rsa XXXXXX
EOD

	${EDITOR:-nano} ${SSH_AUTHKEYS}
		
	# Touch up permissions
	
	chown -R "${USERNAME}:${USERNAME}" "${HOMEDIR}"
	chmod 600 "${SSH_AUTHKEYS}"
	chmod 700 "${BACKUPDATA_DIR}"
	chmod 700 "${HOMEDIR}/.ssh"
		
	# Install attic for user
	
	echo "Installing attic for user..."
	su "${USERNAME}" -c "cd \"${HOMEDIR}\" ; pip-3.2 install -q --user attic"
	
	# Send unix username
	
	echo "Account created. Send this connection code back to the client:"
	echo "  ${USERNAME}"
	
	# Done
}

install_client() {
	
	# Ensure attic is installed for local user
	local ATTIC=~/.local/bin/attic
	if [[ ! -f "$ATTIC" ]] ; then
	
		check_pip

		echo "Installing dependency..."
		pip-3.2 install -q --user attic
	fi
	
	# Build temporary directory
	if [[ -d "$STASHDIR" ]] ; then
		fatal "Backup system already installed! Try --uninstall-client first."
	fi
	
	mkdir "${STASHDIR}"
	mkdir "${STASHDIR}/logs"
	touch "${STASHDIR}/logs/all-jobs"
	
	mkdir -p ~/.ssh
	chmod -R 700 ~/.ssh
	
	# Generate SSH keypair
	(
		echo 25
		ssh-keygen -q -f "${STASHDIR}/repository.rsa" -t rsa -b 4096 -N ''
		echo 100
		sleep 1
	) | whiptail --gauge 'Generating encryption key...' 10 60 0
	
	# Display keypair for copying
	#whiptail --textbox "${STASHDIR}/repository.rsa.pub" 20 75 --title 'Copy this code to your support contact'
	clear
	echo "Copy this code to your support contact:"
	echo ""
	cat "${STASHDIR}/repository.rsa.pub"
	echo ""
	echo "Press enter to continue..."
	read
		
	# Read remote connection detail
	whiptail --inputbox "Enter the connection code from your support contact:" 10 60 2>"${STASHDIR}/remoteuser"
	
	# Read remote connection detail
	whiptail --inputbox "Enter the backup server address from your support contact:" 10 60 2>"${STASHDIR}/remotehost"
	
	# Build entry in .ssh/config
	BACKUPSYS="BSHOST$(date +%s)"
	REMOTEUSER="$(cat ${STASHDIR}/remoteuser)"
	
	cat <<EOD >> ~/.ssh/config
	
Host $BACKUPSYS
	HostName $(cat ${STASHDIR}/remotehost)
	Port 22
	User ${REMOTEUSER}
	IdentityFile ${STASHDIR}/repository.rsa
	
EOD
	
	# Prompt for making any changes to .ssh/config
	if whiptail --yesno "Do you need to make any changes to the backup server address?" 10 60 ; then
		${EDITOR:-nano} ~/.ssh/config
	fi
	
	# Build repository passphrase
	dd if=/dev/urandom "of=${STASHDIR}/passphrase.raw" bs=4096 count=1 2>/dev/null
	base64 -w 0 "${STASHDIR}/passphrase.raw" > "${STASHDIR}/passphrase"
	
	# Init remote repository
	echo "Initializing remote repository, please wait..."
	
	export ATTIC_PASSPHRASE="$(cat ${STASHDIR}/passphrase)"	
	REPOCMD="${BACKUPSYS}:/home/${REMOTEUSER}/backupdata/repo.attic"
	$ATTIC init --encryption=passphrase "${REPOCMD}"
	
	# Build backup command
	
	cat <<EOD >${STASHDIR}/runbackup.sh
#!/bin/bash

REPOSITORY="${REPOCMD}"
BACKUPJOB="\$(date '+%Y-%m-%d-%H-%M-%S')"
LOGFILE="${STASHDIR}/logs/\${BACKUPJOB}.log"
export ATTIC_PASSPHRASE="\$(cat ${STASHDIR}/passphrase)"
ATTIC=~/.local/bin/attic

date >> \${LOGFILE}

# {{
# Backup tasks go here

# e.g.
\$ATTIC create --stats "\${REPOSITORY}::my-home-directory-\${BACKUPJOB}" /home/$(whoami) >> \${LOGFILE}

# }}}

# Prune old backup jobs

echo "Pruning..." >> \${LOGFILE}
\$ATTIC prune -s --keep-within=180d "\${REPOSITORY}" >> \${LOGFILE}

# Keep track of current backup jobs

echo "Updating all-jobs..." >> \${LOGFILE}
\$ATTIC list "\${REPOSITORY}" > "${STASHDIR}/logs/all-jobs"

# Done

date >> \${LOGFILE}

EOD
	chmod +x "${STASHDIR}/runbackup.sh"
	
	# Prompt for easy-password, for keyphrase recovery
	EASYPASS=$(mktemp)
	whiptail --passwordbox 'Enter a password to back up encryption keys to remote server.\nWithout a backup of the keys, all data could be lost!\n\nLeave password field blank to not back up configuration.' 10 75 2>"${EASYPASS}"
		
	if [[ -s "${EASYPASS}" ]] ; then
		
		ENCFILE=$(mktemp)
		openssl enc -aes-256-cbc -salt -in "${STASHDIR}/passphrase.raw" -out "${ENCFILE}" -k "$(cat ${EASYPASS})"
		
		rm "${TARFILE}"
		
		# Remote site blocks scp, this has to be manual
		
		clear
		echo "Backup starts here:"
		echo ""
		base64 "${ENCFILE}"
		echo ""
		echo ""
		echo "Scroll up, and copy the whole section to your support contact!"
		echo ""
		echo "Press enter to continue..."
		read
		
		rm "${ENCFILE}"
		
	fi
	
	rm "${EASYPASS}"
	
	# All done
	
	clear	
	cat <<EOD
Attic has been installed!

To configure your backup sources and retention policies:
   Edit the file ${STASHDIR}/runbackup.sh

To back up automatically on schedule:
   Run "crontab -e" and reference the above

To restore files:
   Copy the above file and modify it to use "attic mount".
   
EOD
}

uninstall_client() {
	
	if [[ ! -d "$STASHDIR" ]] ; then
		echo "Backup system not installed!" >&2
		exit 1
		
	elif ! whiptail --yesno 'Keep the backup system and all keys?' 10 60 ; then
		rm -r "$STASHDIR"
		echo 'Backup system and all keys removed!'
		# We've left a little mess behind in .ssh/config. But it won't interfere
		#  with future installs as it is datestamped.
		
		# .
		
	#else
		# Nothing to do, client didn't confirm removal
		
	fi
}

usage() {
	cat >&2 <<EOD 
Usage: atticinst [options]
 --client            Install as client
 --server            Install as server.
                       Optionally accepts ATTICINST_USERNAME as an environment
                       variable to control the generated username.
 --dependencies      Install dependencies only
 --uninstall-client  Uninstall client from current user
EOD
	
	exit 1
}

main() {

	case "$1" in
		"--dependencies")
			install_pip
		;;
	
		"--uninstall-client")
			uninstall_client
		;;
	
		"--client")
			install_client
		;;

		"--server")
			install_server
		;;
		
		*)
			usage
		;;
	esac

	return 0
}

main "$@"
