#!/bin/bash

set -e

ATTIC_EXPECT="Attic 0.13"
ATTIC_BINARY_64="https://attic-backup.org/downloads/releases/0.13/Attic-0.13-linux-x86_64.tar.gz"
ATTIC_HASH_64="c667d97df54ffb6037bd1f326577b42a"
ATTIC_BINARY_32="https://attic-backup.org/downloads/releases/0.13/Attic-0.13-linux-i686.tar.gz"
ATTIC_HASH_32="c8d0514b4b60bee52bc8982ff16c561c"

MUNIN_PLUGIN_PREFIX=/etc/munin/plugins/attic_

get_stashdir() {
	if [[ -n "$ATTICINST_DIR" ]] ; then
		echo "$ATTICINST_DIR"
	else
		# Default stash directory
		echo "${HOME}/.atticman"
	fi
}

fatal() {
	echo "FATAL: $1" >&2
	exit 1
}

require_root() {
	if [[ "$UID" != "0" ]] ; then
		fatal "This operation requires root!"
	fi
}

package_install() {
	require_root

	if ! dpkg -s "$1" > /dev/null 2>&1 ; then
		apt-get -y install "$1"
	fi
}

package_check() {
	if ! dpkg -s "$1" > /dev/null 2>&1 ; then
		fatal "Missing dependency '{$1}' - please run --dependencies as root"
	fi
}

dependencies_install() {
	package_install whiptail
	package_install openssl
	package_install wget
	package_install bc
}

dependencies_check() {
	package_check whiptail
	package_check openssl
	package_check wget
	package_check bc
}

download_attic() {
	# Usage: download_attic {destination-path}
	
	local ATTIC_BINARY=""
	local ATTIC_HASH=""
	
	# Determine host type	
	case "${HOSTTYPE}" in
		x86_64)
			ATTIC_BINARY="${ATTIC_BINARY_64}"
			ATTIC_HASH="${ATTIC_HASH_64}"
		;;
		
		i386|i486|i586|i686)
			ATTIC_BINARY="${ATTIC_BINARY_32}"
			ATTIC_HASH="${ATTIC_HASH_32}"
		;;
		
		*)
			fatal "Unsupported architecture '${HOSTTYPE}'"
		;;
	esac
	
	# Download
	local ATTICTAR=$(mktemp)
	wget -O "${ATTICTAR}" "${ATTIC_BINARY}"
	
	# Check hash
	DLHASH=$(md5sum "${ATTICTAR}" | cut -d' ' -f1)	
	if [[ ! "${DLHASH}" = "${ATTIC_HASH}" ]] ; then
		fatal "Downloaded attic binary doesn't match expected hash!"
	fi
	
	# Extract
	tar -x -C "$1" --strip-components=1 -f "${ATTICTAR}"
	rm "${ATTICTAR}"
	
	# Verify banner
	if ! "${1}/attic" help | grep -q "${ATTIC_EXPECT}" ; then
		fatal "Missing expected canary in attic banner!"
	fi
}

install_server() {
	
	dependencies_check
	
	# Create user account
	
	echo "Creating system user..."
	
	local USERNAME="AB$(date +%s)"
	if [[ -n "$ATTICINST_USERNAME" ]] ; then
		USERNAME="$ATTICINST_USERNAME"
	fi
	
	local PASSWORD=$( ( dd if=/dev/urandom count=1 bs=256 2>/dev/null) | md5sum -b | cut -d' ' -f1)
	local HOMEDIR=/home/${USERNAME}
	
	useradd "$USERNAME"
	echo "${USERNAME}:${PASSWORD}" | chpasswd
	
	# Create folder structure
	
	mkdir -p "${HOMEDIR}/backupdata"
	mkdir -p "${HOMEDIR}/.ssh"
	
	# Read ssh pubkey
	
	local AUTHKEYFILE="${HOMEDIR}/.ssh/authorized_keys"
	cat <<EOD > ${AUTHKEYFILE}
# Paste in the public key from the client...
command="${HOMEDIR}/connection-wrapper",no-port-forwarding,no-X11-forwarding,no-agent-forwarding ssh-rsa XXXXXX
EOD

	${EDITOR:-nano} ${AUTHKEYFILE}
		
	# Install attic for user
	
	echo "Installing attic..."
	mkdir -p "${HOMEDIR}/bin"
	download_attic "${HOMEDIR}/bin"
	
	# Create wrapper logging script
	
	echo <<EOD > "${HOMEDIR}/connection-wrapper"
#!/bin/bash
echo "\$(date +%s) \${SSH_CONNECTION} start" >> "${HOMEDIR}/connection-log"
${HOMEDIR}/bin/attic serve
echo "\$(date +%s) \${SSH_CONNECTION} stop" >> "${HOMEDIR}/connection-log"
EOD
	
	# Touch up permissions
	
	chown -R "${USERNAME}:${USERNAME}" "${HOMEDIR}"
	chmod 600 "${AUTHKEYFILE}"
	chmod 700 "${HOMEDIR}/backupdata"
	chmod 700 "${HOMEDIR}/.ssh"
	
	chmod 755 "${HOMEDIR}/bin/attic"
	
	chmod 755 "${HOMEDIR}/connection-wrapper"
	
	touch "${HOMEDIR}/connection-log"
	chmod 640 "${HOMEDIR}/connection-log"
	
	# Send unix username
	
	echo "Account created. Send this connection code back to the client:"
	echo "  ${USERNAME}"
	
	# Done
}

generate_keypair() {
	(
		echo 25
		ssh-keygen -q -f "$1" -t rsa -b 4096 -N ''
		echo 100
		sleep 1
	) | whiptail --gauge 'Generating encryption key...' 10 60 0
}

install_client() {
	
	local STASHDIR=$(get_stashdir)
	
	# Build temporary directory
	if [[ -d "$STASHDIR" ]] ; then
		fatal "Backup system already installed! Try --uninstall-client first."
	fi
	
	mkdir "${STASHDIR}"
	mkdir "${STASHDIR}/logs"
	mkdir "${STASHDIR}/attic"
	mkdir -p ~/.ssh
	
	touch "${STASHDIR}/logs/all-jobs"
	chmod -R 700 ~/.ssh
	
	download_attic "${STASHDIR}/attic"
	
	# Generate SSH keypair
	generate_keypair "${STASHDIR}/repository.rsa"
	
	# Display keypair for copying
	clear
	echo "Copy this code to your support contact:"
	echo ""
	cat "${STASHDIR}/repository.rsa.pub"
	echo ""
	echo "Press enter to continue..."
	read
		
	# Read remote connection detail
	whiptail --inputbox "Enter the connection code from your support contact:" 10 60 2>"${STASHDIR}/remoteuser"
	local REMOTEUSER=$(cat "${STASHDIR}/remoteuser")
	rm "${STASHDIR}/remoteuser"
	
	[[ "${REMOTEUSER}" = "" ]] && fatal "Connection code not entered."
	
	# Read remote connection detail
	whiptail --inputbox "Enter the backup server address from your support contact:" 10 60 2>"${STASHDIR}/remotehost"
	local REMOTEHOST=$(cat "${STASHDIR}/remotehost")
	rm "${STASHDIR}/remotehost"
	
	[[ "${REMOTEUSER}" = "" ]] && fatal "Remote server address not entered."
	
	local REMOTEPORT=22
	
	local regex="(.*):(.*)"
	if [[ $REMOTEHOST =~ $regex ]] ; then
		REMOTEHOST="${BASH_REMATCH[1]}"
		REMOTEPORT="${BASH_REMATCH[2]}"
	fi
	
	# Build entry in .ssh/config
	BACKUPSYS="BSHOST$(date +%s)"
	
	cat <<EOD >> ~/.ssh/config
	
Host $BACKUPSYS
	HostName ${REMOTEHOST}
	Port ${REMOTEPORT}
	User ${REMOTEUSER}
	IdentityFile ${STASHDIR}/repository.rsa
	
EOD
	
	echo "${BACKUPSYS}" > "${STASHDIR}/remotesite"
	
	# Build repository passphrase
	
	if whiptail --yesno 'Use automatically generated passphrase?' 10 60 ; then
		dd if=/dev/urandom "of=${STASHDIR}/passphrase.raw" bs=512 count=1 2>/dev/null
		base64 -w 0 "${STASHDIR}/passphrase.raw" > "${STASHDIR}/passphrase"
		rm "${STASHDIR}/passphrase.raw"
	else
		whiptail --inputbox "Enter encryption passphrase:" 10 60 2>"${STASHDIR}/passphrase"
	fi
	
	# Init remote repository
	echo "Initializing remote repository, please wait..."
	
	export ATTIC_PASSPHRASE="$(cat ${STASHDIR}/passphrase)"	
	local REPOCMD="${BACKUPSYS}:/home/${REMOTEUSER}/backupdata/repo.attic"
	${STASHDIR}/attic/attic init --encryption=passphrase "${REPOCMD}"
	
	# Build backup command
	
	cat <<EOD >${STASHDIR}/runbackup.sh
#!/bin/bash

REPOSITORY="${REPOCMD}"
BACKUPJOB="\$(date '+%Y-%m-%d-%H-%M-%S')"
LOGFILE="${STASHDIR}/logs/\${BACKUPJOB}.log"
export ATTIC_PASSPHRASE="\$(cat ${STASHDIR}/passphrase)"
ATTIC="${STASHDIR}/attic/attic"

backuptasks() {
	# Backup tasks go here
	# e.g.
	\$ATTIC create --stats "\${REPOSITORY}::my-home-directory-\${BACKUPJOB}" /home/$(whoami)
}

log_inside() {
	date
	
	echo "Starting backup..."	
	backuptasks

	echo "Pruning old jobs..."
	\$ATTIC prune -s --keep-within=180d "\${REPOSITORY}"

	echo "Updating all-jobs..."
	\$ATTIC list "\${REPOSITORY}" > "${STASHDIR}/logs/all-jobs"

	date
}

log_inside > \${LOGFILE}
EOD
	chmod +x "${STASHDIR}/runbackup.sh"
	
	# Copy installer into stashdir
	
	cp "${BASH_SOURCE[0]}" "${STASHDIR}/atticinst.sh"
	chmod 755 "${STASHDIR}/atticinst.sh"
	
	# Backup client configuration
	
	backup_client_configuration
	
	# All done
	
	clear	
	cat <<EOD
Attic has been installed!

To configure your backup sources and retention policies:
   Edit the file ${STASHDIR}/runbackup.sh

To back up automatically on schedule:
   Run "crontab -e" and reference the above

To restore files:
   Copy the above file and modify it to use "attic mount".
   
EOD
}

backup_client_configuration() {

	local STASHDIR=$(get_stashdir)
	
	if [[ ! -f "${STASHDIR}/passphrase" ]] ; then
		fatal "Client not installed"
	fi

	# Prompt for easy-password, for keyphrase recovery
	local EASYPASS=$(mktemp)
	whiptail --passwordbox 'Enter a password to back up encryption keys to remote server.\nWithout a backup of the keys, all data could be lost!\n\nLeave password field blank to not back up configuration.' 10 75 2>"${EASYPASS}"
		
	if [[ -s "${EASYPASS}" ]] ; then
		
		local ENCFILE=$(mktemp)
		openssl enc -aes-256-cbc -salt -in "${STASHDIR}/passphrase" -out "${ENCFILE}" -k "$(cat ${EASYPASS})"
		
		# Remote site blocks scp, this has to be manual
		
		clear
		echo "Backup starts here:"
		echo ""
		base64 "${ENCFILE}"
		echo ""
		echo ""
		echo "Scroll up, and copy the whole section to your support contact!"
		echo ""
		echo "Press enter to continue..."
		read
		
		rm "${ENCFILE}"
		
	fi
	
	rm "${EASYPASS}"
	
}

uninstall_client() {
	
	local STASHDIR=$(get_stashdir)
	
	if [[ ! -d "${STASHDIR}" ]] ; then
		echo "Backup system not installed!" >&2
		exit 1
		
	elif ! whiptail --yesno 'Keep the backup system and all keys?' 10 60 ; then
		
		echo -n "About to remove directory '${STASHDIR}'... 3..."
		sleep 1
		echo -n "2... "
		sleep 1
		echo -n "1... "
		sleep 1
		echo " "
	
		rm -r "$STASHDIR"
		echo 'Backup system and all keys removed!'
		
		# We've left a little mess behind in .ssh/config. But it won't interfere
		#  with future installs as it is datestamped.
		
	#else
		# Nothing to do, client didn't confirm removal
		
	fi
}

check_stashdir() {
	local STASHDIR=$(get_stashdir)

	# echo "Metadata directory is '${STASHDIR}'"
	
	if [[ -d "${STASHDIR}" ]] ; then
		exit 0
	else
		exit 1
	fi
}

human_to_raw() {
	local input=$(cat)
	local base=$(echo "$input" | cut -d' ' -f1)
	local unit=$(echo "$input" | cut -d' ' -f2)
	
	case "$unit" in
		KB|kb)
			echo "$base * 1024" | bc
		;;
		MB|mb)
			echo "$base * 1024 * 1024" | bc
		;;
		GB|gb)
			echo "$base * 1024 * 1024 * 1024" | bc
		;;
		TB|tb)
			echo "$base * 1024 * 1024 * 1024 * 1024" | bc
		;;
		*)
			echo "$base"
		;;
	esac
}

munin_helper_laststats() {
	# Usage: munin_helper_laststats
	local STASHDIR=$(get_stashdir)
	local LAST_LOGFILE=$(ls "${STASHDIR}/logs" | tail -n2 | head -n1)
	grep 'All archives' "${STASHDIR}/logs/${LAST_LOGFILE}" | tail -n1 | sed -r 's/\s\s+/_/g'
}

munin_stored() {
	if [[ "$1" == "config" ]] ; then
		cat <<EOF
graph_title Deduplicated size
graph_category Attic
graph_args --base 1024 -l 0
ddsize.label Deduplicated size
EOF
	else
		local DDUP_SIZE=$(munin_helper_laststats | cut -d_ -f4 | human_to_raw)
		
		cat <<EOF
ddsize.value $DDUP_SIZE
EOF
	fi
}

munin_since() {	
	if [[ "$1" == "config" ]] ; then
		cat <<EOF
graph_title Time since last cycle
graph_category Attic
timesincelast.label Hours
EOF
	else
		local LASTJOB=$(tail -n1 "$(get_stashdir)/logs/all-jobs" | sed -r 's/^[^ ]+ *//')
		local LASTJOB_UNIX=$(date -d "${LASTJOB}" +%s)
		local LASTJOB_SINCE=$(( $(date +%s) - $LASTJOB_UNIX ))
		local LASTJOB_SINCE_HOURS=$( echo "scale=2; $LASTJOB_SINCE / 3600" | bc)
		
		cat <<EOF
timesincelast.value $LASTJOB_SINCE_HOURS
EOF
	fi
}

munin_totaljobs() {
	if [[ "$1" == "config" ]] ; then
		cat <<EOF
graph_title Total backup archives
graph_category Attic
totaljobs.label Archives
EOF
	else
		local NUMJOBS=$(wc -l < "$(get_stashdir)/logs/all-jobs")
		
		cat <<EOF
totaljobs.value $NUMJOBS
EOF
	fi
}

munin_compression() {
	if [[ "$1" == "config" ]] ; then
		cat <<EOF
graph_title Compression ratio
graph_category Attic
compression.label Percentage
EOF
	else	
		local LAST_STATS=$(munin_helper_laststats)		
		local ORIG_SIZE=$(echo "$LAST_STATS" | cut -d_ -f2 | human_to_raw)
		local COMP_SIZE=$(echo "$LAST_STATS" | cut -d_ -f3 | human_to_raw)
		local COMPRESSION_RATIO=$(echo "scale=2; 100 * $COMP_SIZE / $ORIG_SIZE" | bc)
		
		cat <<EOF
compression.value $COMPRESSION_RATIO
EOF
	fi
}

munin_currentprocs() {
	if [[ "$1" == "config" ]] ; then
		cat <<EOF
graph_title Current Attic processes
graph_category Attic
numprocs.label Processes
EOF
	else
		local CURRENT_PROCESSES=$(ps -C attic | tail -n +2 | wc -l)
		cat <<EOF
numprocs.value $CURRENT_PROCESSES
EOF
	fi
}

munin_node_install() {

	# Create wrapper scripts so that we can explicitly set log directory
	# Munin user must have rx to atticinst, rx to stashdir+logdir, r to logs
	
	local STASHDIR=$(get_stashdir)
	#local SCRIPT_LOCATION="${BASH_SOURCE[0]}"
	local SCRIPT_LOCATION="${STASHDIR}/atticinst.sh"
	
	for arg in stored since totaljobs compression currentprocs ; do
		echo "#!/bin/bash" > ${MUNIN_PLUGIN_PREFIX}$arg
		echo "ATTICINST_DIR=\"$STASHDIR\" $SCRIPT_LOCATION --munin-$arg \"\$@\"" >> ${MUNIN_PLUGIN_PREFIX}$arg
		chmod 755 ${MUNIN_PLUGIN_PREFIX}$arg
	done
	
	service munin-node restart
}

munin_node_uninstall() {
	rm /etc/munin/plugins/attic_{stored,since,totaljobs,compression,currentprocs}
	
	service munin-node restart
}

usage() {
	cat >&2 <<EOD 
Usage:
  atticinst [options..]

Options:
 --client            Install as client
 --server            Install as server
                       Optionally accepts ATTICINST_USERNAME as an environment
                       variable to control the generated username.
 --dependencies      Install dependencies only
 --uninstall-client  Uninstall client from current user
 --backup-config     Re-prompt to backup client configuration
 --is-installed      Check for existence of metadata directory, sets 0/1 error
 --version           Display md5sum of installer script
 
 --munin-install     Install client plugins for munin-node
 --munin-uninstall   Uninstall client plugins for munin-node 
 --munin-{command} ["config"]  Run munin command
 
Directories:
 ~/.atticman/        Metadata directory (override with ATTICINST_DIR)

EOD
	
	exit 1
}

main() {
	case "$1" in
		"--dependencies")
			dependencies_install
		;;
		
		"--uninstall-client")
			uninstall_client
		;;
		
		"--client")
			install_client
		;;
		
		"--server")
			install_server
		;;
		
		"--backup-config")
			backup_client_configuration
		;;
		
		"--is-installed")
			check_stashdir
		;;
		
		"--version")
			md5sum "${BASH_SOURCE[0]}" | cut -d' ' -f1
		;;
		
		"--munin-install")
			munin_node_install
		;;
		
		"--munin-uninstall")
			munin_node_uninstall
		;;
		
		"--munin-stored")
			munin_stored "$2"
		;;
		
		"--munin-since")
			munin_since "$2"
		;;
		
		"--munin-totaljobs")
			munin_totaljobs "$2"
		;;
		
		"--munin-compression")
			munin_compression "$2"
		;;
		
		"--munin-currentprocs")
			munin_currentprocs "$2"
		;;
		
		*)
			usage
		;;
	esac
	
	return 0
}


main "$@"
